using System;
using System.Collections.Generic;
using System.Text;

namespace Labyrinth
{
    /// <summary>
    /// Find a way through the labyrinth.
    /// </summary>
    class LabPathFinder
    {
        class RoomInfo
        {
            public int i, j;
            public byte NextDirection;

            public RoomInfo(int i, int j) : this(i, j, 0) { }
            public RoomInfo(int i, int j, byte nextRoom)
            {
                this.i = i;
                this.j = j;
                this.NextDirection = nextRoom;
            }
        }

        Labyrinth lab;

        /// <summary>
        /// Get or set the problem labyrinth.
        /// </summary>
        public Labyrinth Labyrinth
        {
            get { return lab; }
            set { lab = value; }
        }

        private bool unmark = true;

        /// <summary>
        /// Get or set the "unmark" flag.
        /// If set to true, rooms will be unmarked before path finding.
        /// </summary>
        public bool Unmark
        {
            get { return unmark; }
            set { unmark = value; }
        }

        public LabPathFinder() { }

        public LabPathFinder(Labyrinth lab)
        {
            this.lab = lab;
        }

        /// <summary>
        /// Find way through the labyrinth via depth first search (DFS).
        /// </summary>
        public void FindWayDFS()
        {
            if(unmark)
                lab.UnmarkRooms();
            
            // start DFS with start coordinates
            dfs(lab.StartCoordinates[0], lab.StartCoordinates[1]);
        }

        private bool dfs(int i, int j)
        {
            // mark room as visited
            lab.Rooms[i, j].SetVisited();

            // if end position - return true
            if (i == lab.EndCoordinates[0] && j == lab.EndCoordinates[1])
            {
                return true;
            }

            bool wayFound = false;
            // search recursively for next room in order East -> South -> West -> North
            if (!lab.Rooms[i, j].East && i + 1 < lab.Rooms.GetLength(0) && lab.Rooms[i + 1, j].IsUnvisited())
            {
                wayFound = dfs(i + 1, j);
            }
            if (!wayFound && !lab.Rooms[i, j].South && j + 1 < lab.Rooms.GetLength(1) && lab.Rooms[i, j + 1].IsUnvisited())
            {
                wayFound = dfs(i, j + 1);
            }
            if (!wayFound && !lab.Rooms[i, j].West && i - 1 >= 0 && lab.Rooms[i - 1, j].IsUnvisited())
            {
                wayFound = dfs(i - 1, j);
            }
            if (!wayFound && !lab.Rooms[i, j].North && j - 1 >= 0 && lab.Rooms[i, j - 1].IsUnvisited())
            {
                wayFound = dfs(i, j - 1);
            }

            // if way found -> return true
            if (wayFound)
            {
                return true;
            }
            else
            {
                // if no way found, set room to visited, but not in current path from start to end
                lab.Rooms[i, j].SetVisitedNoPath();
                return false;
            }
        }

        /// <summary>
        /// Find way through the labyrinth via iterative version of depth first search (DFS).
        /// </summary>
        public void FindWayDFSIt()
        {
            if (unmark)
                lab.UnmarkRooms();

            // hold room stack as equivalent to the recursive calls
            Stack<RoomInfo> roomStack = new Stack<RoomInfo>();
            roomStack.Push(new RoomInfo(lab.StartCoordinates[0], lab.StartCoordinates[1]));

            // loop until the end coordinates are reached
            while (true)
            {
                // get current room from the stack
                RoomInfo ri = roomStack.Pop();
                int i = ri.i;
                int j = ri.j;
                
                // mark room as visited
                lab.Rooms[i, j].SetVisited();

                // if the end coordinates are reached -> return
                if (i == lab.EndCoordinates[0] && j == lab.EndCoordinates[1])
                    return;

                // search for next room in order East -> South -> West -> North
                // if the way is free, save the neighbour
                RoomInfo neighbour = null;
                if (ri.NextDirection==0 && !lab.Rooms[i, j].East && i + 1 < lab.Rooms.GetLength(0) && lab.Rooms[i + 1, j].IsUnvisited())
                {
                    neighbour = new RoomInfo(i + 1, j);
                    ri.NextDirection = 1;
                }
                else if (ri.NextDirection<=1 && !lab.Rooms[i, j].South && j + 1 < lab.Rooms.GetLength(1) && lab.Rooms[i, j + 1].IsUnvisited())
                {
                    neighbour = new RoomInfo(i, j + 1);
                    ri.NextDirection = 2;
                }
                else if (ri.NextDirection<=2 && !lab.Rooms[i, j].West && i - 1 >= 0 && lab.Rooms[i - 1, j].IsUnvisited())
                {
                    neighbour = new RoomInfo(i - 1, j);
                    ri.NextDirection = 3;
                }
                else if (!lab.Rooms[i, j].North && j - 1 >= 0 && lab.Rooms[i, j - 1].IsUnvisited())
                {
                    neighbour = new RoomInfo(i, j - 1);
                }
                else
                {
                    lab.Rooms[i, j].SetVisitedNoPath();
                }

                // if neighbour != null  --> push it with the current room on the stack
                if (neighbour != null)
                {
                    roomStack.Push(ri);
                    roomStack.Push(neighbour);
                }
            }
        }

        /// <summary>
        /// Finds the way via "turn counter" method ( = Pledge-algorithm ).
        /// http://www-i1.informatik.rwth-aachen.de/~algorithmus/algo6.php
        /// </summary>
        public void FindWayTurnCounter()
        {
            if(unmark)
                lab.UnmarkRooms();
            
            // set current room to the start coordinates
            int i = lab.StartCoordinates[0];
            int j = lab.StartCoordinates[1];
            int direction = 0;   // direction: 0=east, 1=south, 2=west, 3=north

            int counter = 0;
            // while not end position - go one step forward
            while (i != lab.EndCoordinates[0] || j != lab.EndCoordinates[1])
            {
                // set current room to visited
                lab.Rooms[i, j].SetVisited();

                // if no wall reached -> one step into current directon
                if (counter == 0)
                {
                    if (wallInDirection(i, j, direction))       // if wall reached -> turn right
                    {
                        direction = (direction + 1) % 4;
                        counter--;
                    }
                    else                                        // else update position
                    {
                        i += xdiff(direction);
                        j += ydiff(direction);
                    }
                }
                else                                            // else go along the wall until counter==0
                {
                    direction--;                                // turn left if possible
                    if (direction == -1)                        // trick for -1 % 4    (= 3)
                        direction = 3;
                    counter++;
                    while (wallInDirection(i, j, direction))    // else search for next possible direction
                    {
                        direction = (direction + 1) % 4;
                        counter--;
                    }
                    i += xdiff(direction);                      // make one step
                    j += ydiff(direction);
                }
            }

            // set end coordinates to visited
            lab.Rooms[lab.EndCoordinates[0], lab.EndCoordinates[1]].SetVisited();
        }

        /// <summary>
        /// Calculates the difference of the x coordinate from the given direction.
        /// </summary>
        /// <param name="dir">current direction</param>
        /// <returns>x difference</returns>
        private int xdiff(int dir)
        {
            return (dir == 0) ? 1 : (dir == 2) ? -1 : 0;
        }

        /// <summary>
        /// Calculates the difference of the y coordinate from the given direction.
        /// </summary>
        /// <param name="dir">current direction</param>
        /// <returns>y difference</returns>
        private int ydiff(int dir)
        {
            return (dir == 1) ? 1 : (dir == 3) ? -1 : 0;
        }

        /// <summary>
        /// Checks if the room has a wall in the given direction.
        /// </summary>
        /// <param name="i">Room index 0</param>
        /// <param name="j">Room index 1</param>
        /// <param name="dir">current direction</param>
        /// <returns>true, if the room [i,j] has a wall in the given direction</returns>
        private bool wallInDirection(int i, int j, int dir)
        {
            // if we reach the start coordinates -> north has a wall!
            if (j == 0 && dir == 3)
                return true;

            // check if there's a wall in the given direction
            if (dir == 0)
                return lab.Rooms[i, j].East;
            else if (dir == 1)
                return lab.Rooms[i, j].South;
            else if (dir == 2)
                return lab.Rooms[i, j].West;
            else
                return lab.Rooms[i, j].North;
        }
    }
}
