using System;
using System.Collections.Generic;
using System.Text;

namespace Labyrinth
{
    class Labyrinth
    {
        //stuff to create the labyrinth:

        //list of roomNumbers in a UnionFind wrapper
        //- each element is assigned to a set
        private UnionFind<int>[] roomNumbers;

        //list of walls
        //each wall is defined by the room numbers of the rooms that it seperates
        //first index: the wall
        //second index: room 0 or 1
        private int[][] walls;

        private Random random = new Random();

        private Room[,] rooms;

        public Room[,] Rooms
        {
            get { return rooms; }
            set { rooms = value; }
        }

        private int[] startCoordinates = new int[] { 0, 0 };
        /// <summary>
        /// Coordinates in the Rooms matrix giving the entrance point to the labyrinth
        /// </summary>
        public int[] StartCoordinates
        {
            get { return startCoordinates; }
        }

        private int[] endCoordinates = null;
        /// <summary>
        /// Coordinates in the Rooms matrix giving the exit point from the labyrinth
        /// </summary>
        public int[] EndCoordinates
        {
            get { return endCoordinates; }
        }

        /// <summary>
        /// Creates a labyrinth of the size nxn.
        /// </summary>
        /// <param name="n">width and height</param>
        /// <param name="looseness">0 means there is only one way through the labyrinth.
        ///                         Values up to 100 lead to more ways.</param>
        public Labyrinth(int n, byte looseness)
        {
            if (looseness > 100) looseness = 100;
            roomNumbers = new UnionFind<int>[n * n];
            walls = new int[n * (n - 1) * 2][];

            //create the rooms
            for (int i = 0; i < roomNumbers.Length; i++)
            {
                roomNumbers[i] = new UnionFind<int>(i, null);
            }

            //define all the walls
            for (int i = 0; i < walls.Length; i++)
            {
                if (i < n * (n - 1))
                {
                    //vertical walls
                    walls[i] = new int[] { i + (i + i / n + 1) / n, i + (i + i / n + 1) / n + 1 };
                }
                else
                {
                    //horizontal walls
                    walls[i] = new int[] { i - n * (n - 1), i - n * (n - 1) + n };
                }

            }

            //delete walls to create ways
            for (int i = 0; i < walls.Length; i++)
            {
                int k = random.Next(i, walls.Length);
                int[] tmp = walls[i];
                walls[i] = walls[k];
                walls[k] = tmp;
                if (roomNumbers[walls[i][0]].Set != roomNumbers[walls[i][1]].Set || random.Next(100) < looseness)
                {
                    roomNumbers[walls[i][0]].Union(roomNumbers[walls[i][1]]);
                    walls[i] = null;
                }
            }

            //convert the labyrinth from wall-list form to room matrix:

            //create rooms
            rooms = new Room[n, n];
            for (int i = 0; i < n; i++)
            {
                for (int j = 0; j < n; j++)
                {
                    rooms[i, j] = new Room();
                    if (i == 0) rooms[i, j].West = true;
                    if (i == n - 1) rooms[i, j].East = true;
                    if (j == 0) rooms[i, j].North = true;
                    if (j == n - 1) rooms[i, j].South = true;
                }
            }
            rooms[0, 0].North = false;
            rooms[n - 1, n - 1].South = false;

            // set exit point
            this.endCoordinates = new int[] { n - 1, n - 1 };

            //set the walls
            foreach (int[] w in walls)
            {
                if (w != null)
                {
                    if (w[1] - w[0] == 1)
                    {
                        //vertical wall
                        rooms[w[0] % n, w[0] / n].East = true;
                        rooms[w[1] % n, w[1] / n].West = true;
                    }
                    else
                    {
                        //horizontal wall
                        rooms[w[0] % n, w[0] / n].South = true;
                        rooms[w[1] % n, w[1] / n].North = true;
                    }
                }
            }
        }

        /// <summary>
        /// Creates an empty Labyrinth instance for cloning
        /// </summary>
        private Labyrinth() { }

        /// <summary>
        /// Clones the Labyrinth. Each Room is copied and not just the reference.
        /// You may specifiy if the mark of the rooms will be copied, too or if it will have its default value.
        /// </summary>
        /// <param name="cloneMark">If the marks will be copied or ignored</param>
        /// <returns>A labyrinth just like the object you cloned it from</returns>
        public Labyrinth Clone(bool cloneMark)
        {
            Labyrinth l = new Labyrinth();
            l.Rooms = new Room[this.Rooms.GetLength(0), this.Rooms.GetLength(1)];
            for (int i = 0; i < l.Rooms.GetLength(0); i++)
            {
                for (int j = 0; j < l.Rooms.GetLength(1); j++)
                {
                    l.Rooms[i, j] = new Room(
                        this.Rooms[i, j].North,
                        this.Rooms[i, j].East,
                        this.Rooms[i, j].South,
                        this.Rooms[i, j].West);
                    if (cloneMark)
                    {
                        if (l.Rooms[i, j].IsUnvisited()) this.Rooms[i, j].SetUnvisited();
                        if (l.Rooms[i, j].IsVisited()) this.Rooms[i, j].SetVisited();
                        if (l.Rooms[i, j].IsVisitedNoPath()) this.Rooms[i, j].SetVisitedNoPath();
                    }
                }
            }
            l.startCoordinates = new int[] { this.startCoordinates[0], this.startCoordinates[1] };
            l.endCoordinates = new int[] { this.endCoordinates[0], this.endCoordinates[1] };

            return l;
        }

        /// <summary>
        /// Sets all rooms of the labyrinth to unvisited.
        /// </summary>
        public void UnmarkRooms()
        {
            foreach (Room r in this.rooms)
                r.SetUnvisited();
        }

        /// <summary>
        /// Counts the marked rooms (which have been visited before) of the labyrinth.
        /// </summary>
        /// <returns>count of marked rooms</returns>
        public int CountMarkedRooms()
        {
            int count = 0;
            foreach (Room r in this.rooms)
                if (!r.IsUnvisited())
                    count++;
            return count;
        }
    }
}
