package algorithms.priorityqueues;

import java.util.Hashtable;
import java.util.Map;
import java.util.Set;

public class BinomialHeap<K,V> implements PriorityQueue<K,V> {

	BinomialHeapNode<K,V> root = null;
	Map<K,BinomialHeapNode<K,V>> dataAccess = new Hashtable<K,BinomialHeapNode<K,V>>(); 
	
	public BinomialHeap() { }
	public BinomialHeap(BinomialHeapNode<K,V> root){
		this.root = root;
	}
	public BinomialHeap(V[] values, K[] keys, int[] priorities) throws IllegalArgumentException{
		if(values.length!=priorities.length || values.length!=keys.length)
			throw new IllegalArgumentException("Argument arrays must have the same dimension");
		
		for(int i=0; i<values.length; i++){
			BinomialHeapNode<K,V> node = new BinomialHeapNode<K,V>(values[i], keys[i], priorities[i]);
			insert(node);
		}
	}
	
	public Set<K> getKeySet() {
		return dataAccess.keySet();
	}
	
	public BinomialHeapNode<K,V> getNode(K key) {
		return dataAccess.get(key);
	}
	
	public int size(){
		return size(root);
	}
	
	private int size(BinomialHeapNode<K,V> node){
		if(null == node)
			return 0;
		else
			return size(node.getFirstChild()) + size(node.getSibling()) + 1;
			
	}
	
	public boolean checkQueueIntegrity(){
		BinomialHeapNode<K,V> node = root;
		while(null != node){
			if(null!=node.getSibling() && node.getDegree()>=node.getSibling().getDegree())
				return false;
			node = node.getSibling();
		}
		return checkQueueIntegrity(root);
	}
	
	private boolean checkQueueIntegrity(BinomialHeapNode<K,V> node){
		if(null == node)
			return true;
		BinomialHeapNode<K,V> child = node.getFirstChild();
		while(null != child){
			if(child.getPriority() < node.getPriority())
				return false;
			child = child.getSibling();
		}
		return checkQueueIntegrity(node.getFirstChild()) && checkQueueIntegrity(node.getSibling()) && checkDegree(node);
	}
	
	private boolean checkDegree(BinomialHeapNode<K,V> node){
		int deg = 0;
		BinomialHeapNode<K,V> child = node.getFirstChild();
		while(null != child){
			deg++;
			child = child.getSibling();
		}
		return (deg == node.getDegree());
	}
	
	public void decreasePriority(PriorityQueueNode<K,V> node, int newValue) {
		decreasePriority((BinomialHeapNode<K,V>)node, newValue);
	}
	
	private BinomialHeapNode<K,V> decreasePriority(BinomialHeapNode<K,V> n, int newValue) {
		n.setPriority(newValue);
		while(null!=n.getParent() && n.getParent().getPriority()>n.getPriority()){
			swapWithParent(n);
			n = n.getParent();
		}
		return n;
	}
	
	private void swapWithParent(BinomialHeapNode<K,V> n){
		BinomialHeapNode<K,V> p = n.getParent();
		if(null != p){
			// remove p and n from the hashtable
			dataAccess.remove(p.getKey());
			dataAccess.remove(n.getKey());
			
			// swap attributes between the nodes
			n.swapValues(p);
			
			// add p and n with new values to the hashtable
			dataAccess.put(p.getKey(), p);
			dataAccess.put(n.getKey(), n);
		}
	}

	public boolean delete(PriorityQueueNode<K,V> node) {
		decreasePriority(node, Integer.MIN_VALUE);
		deleteMinimum();
		return true;
	}

	public PriorityQueueNode<K,V> deleteMinimum() {
		// get minimum
		BinomialHeapNode<K,V> node = root;
		BinomialHeapNode<K,V> left = null;
		BinomialHeapNode<K,V> minNode = root;
		BinomialHeapNode<K,V> leftMin = null;
		while(node != null){
			if(node.getPriority() < minNode.getPriority()){
				minNode = node;
				leftMin = left;
			}
			left = node;
			node = node.getSibling();
		}
		
		if(null != minNode){
			// delete minNode from root list
			if(minNode == root){
				root = minNode.getSibling();
			}
			else{
				leftMin.setSibling(minNode.getSibling());
			}
			// invert child list and merge with root list
			minNode.invertChildList();
			merge(minNode.getFirstChild());
		}
		
		return minNode;
	}

	public PriorityQueueNode<K,V> getMinimum() {
		BinomialHeapNode<K,V> node = root;
		BinomialHeapNode<K,V> minNode = root;
		while(node != null){
			if(node.getPriority() < minNode.getPriority())
				minNode = node;
			node = node.getSibling();
		}
		return minNode;
	}

	public void insert(PriorityQueueNode<K,V> x) {
		BinomialHeapNode<K,V> x1 = (BinomialHeapNode<K,V>)x;
		merge(x1);
		dataAccess.put(x1.getKey(), x1);
	}

	@SuppressWarnings("unchecked")
	public void merge(PriorityQueue<K,V> q2) {
		BinomialHeap<K,V> bh2 = (BinomialHeap<K,V>)q2;
		merge(bh2.root);
		dataAccess.putAll(bh2.dataAccess);
	}
	
	private void merge(BinomialHeapNode<K,V> q2) {
		this.root = merge(this.root, q2);
		//this.root.setParent(null);
	}
	
	private BinomialHeapNode<K,V> merge(BinomialHeapNode<K,V> a, BinomialHeapNode<K,V> b){
		// join top-lists of a and b with increasing degree
		BinomialHeapNode<K,V> h = joinTopLists(a, b);
		BinomialHeapNode<K,V> hh = h;
		while(null != hh){
			hh.setParent(null);
			hh = hh.getSibling();
		}
		if(null == h)  return h;
		
		BinomialHeapNode<K,V> cur = h;
		BinomialHeapNode<K,V> predec = null;
		BinomialHeapNode<K,V> dec = h.getSibling();
		// watch all elements of the top list
		while(null != dec){
			// if nothing to join or decessor's brother has same degree -> one step further
			if(cur.getDegree()!=dec.getDegree() || 
			  (null!=dec.getSibling() && dec.getSibling().getDegree()==cur.getDegree())){
				predec = cur;
				cur = dec;
			}
			// hang decessor under current node
			else if(cur.getPriority() <= dec.getPriority()){
				cur.setSibling(dec.getSibling());
				dec.linkWith(cur);
			}
			// hang current node under decessor
			else {
				// current node is root of top list -> set new root
				if(null == predec)
					h = dec;
				else
					predec.setSibling(dec);
				cur.linkWith(dec);
				cur = dec;
			}
			dec = cur.getSibling();
		}
		
		return h;
	}

	private BinomialHeapNode<K,V> joinTopLists(BinomialHeapNode<K,V> a, BinomialHeapNode<K,V> b){
		if(null == a)  return b;
		if(null == b)  return a;
		
		// first element of joined top list
		BinomialHeapNode<K,V> h = null;
		BinomialHeapNode<K,V> g = null;
		
		// while both node lists have elements -> add next node to joined top list
		BinomialHeapNode<K,V> cura = a;
		BinomialHeapNode<K,V> curb = b; 
		while(cura!=null && curb!=null){
			// add current node of a
			if(cura.getDegree() <= curb.getDegree()){
				if(null == h){
					h = cura;
					g = h;
				}
				else{
					g.setSibling(cura);
					g = g.getSibling();
				}
				cura = cura.getSibling();
			}
			// add current node of b
			else{
				if(null == h){
					h = curb;
					g = h;
				}
				else{
					g.setSibling(curb);
					g = g.getSibling();
				}
				curb = curb.getSibling();
			}
		}
		
		// one list became empty -> add rest of the other list to joind top list h
		if(null == cura)
			g.setSibling(curb);
		else
			g.setSibling(cura);
		
		return h;
	}
	
}
